import argparse
import enum
import subprocess  # nosec


@enum.unique
class FixtureBase(enum.Enum):
    POSTGRES = "postgres"
    MONGO = "mongo"
    MYSQL = "mysql"

    @classmethod
    def options(cls):
        return ", ".join(fixture_base.value for fixture_base in cls)

    @property
    def command(self):
        postgres_command = [
            "docker",
            "run",
            "-d",
            "--rm",
            "-p",
            "5532:5432",
            "-e",
            "POSTGRES_DB=dev",
            "-e",
            "POSTGRES_USER=user",
            "-e",
            "POSTGRES_PASSWORD=password",
            "postgres:9.6.10-alpine",
        ]
        mongo_command = ["docker", "run", "-d", "--rm", "-p", "28017:27017", "mongo:3.6"]
        mysql_command = [
            "docker",
            "run",
            "-d",
            "--rm",
            "-p",
            "3406:3306",
            "-e",
            "MYSQL_DATABASE=dev",
            "-e",
            "MYSQL_ROOT_PASSWORD=password",
            "mysql:5.6",
        ]
        fixture_base_command_map = {
            FixtureBase.MONGO: mongo_command,
            FixtureBase.POSTGRES: postgres_command,
            FixtureBase.MYSQL: mysql_command,
        }

        return fixture_base_command_map[self]


# TODO: Add an options arg to DOWN a fixture_base's container
parser = argparse.ArgumentParser(
    description="Premptively run docker containers to speed up initial startup of PMR Fixtures."
)
parser.add_argument(
    "fixture_bases",
    metavar="Fixture Base",
    type=str,
    nargs="+",
    help="Available Fixture Bases: {}".format(FixtureBase.options()),
)


def main():
    args = parser.parse_args()

    for fixture_base in args.fixture_bases:
        command = FixtureBase(fixture_base).command
        subprocess.run(command)


if __name__ == "__main__":
    main()



import time

import docker
import responses


class ContainerCheckFailed(Exception):
    """Unable to connect to a Container."""


def get_container(config, ports, environment, check_fn):
    # XXX: moto library may over-mock responses. SEE: https://github.com/spulec/moto/issues/1026
    responses.add_passthru("http+docker")

    def retriable_check_fn(retries):
        while retries:
            retries -= 1
            try:
                check_fn(config)
                return
            except ContainerCheckFailed:
                if not retries:
                    raise
                time.sleep(0.5)

    try:
        container = None
        try:
            retriable_check_fn(1)
        except ContainerCheckFailed:
            try:
                client = docker.from_env(version="auto")
                container = client.containers.run(
                    config.image, ports=ports, environment=environment, detach=True, remove=True
                )
            except docker.errors.APIError as e:
                if "port is already allocated" not in str(e):
                    raise

            retriable_check_fn(40)

        yield
    except Exception:
        raise

    finally:
        if container:
            container.kill()



# flake8: noqa
from pytest_mock_resources.container.mongo import _mongo_container, MongoConfig
from pytest_mock_resources.container.mysql import _mysql_container, MysqlConfig
from pytest_mock_resources.container.postgres import _postgres_container, PostgresConfig
from pytest_mock_resources.container.redis import _redis_container, RedisConfig
from pytest_mock_resources.container.redshift import _redshift_container



import abc
import fnmatch
from typing import Tuple

import attr
import six
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.sql.ddl import CreateSchema
from sqlalchemy.sql.schema import Table

from pytest_mock_resources import compat


@six.add_metaclass(abc.ABCMeta)
class AbstractAction(object):
    @abc.abstractmethod
    def run(self, engine_manager):
        """Run an action on a database via the passed-in engine_manager instance."""


class Rows(AbstractAction):
    def __init__(self, *rows):
        self.rows = rows

    def run(self, engine_manager):
        rows = self._get_stateless_rows(self.rows)

        metadatas = self._get_metadatas(rows)

        for metadata in metadatas:
            engine_manager.create_ddl(metadata)

        self._create_rows(engine_manager.engine, rows)

    @staticmethod
    def _get_stateless_rows(rows):
        """Create rows that aren't associated with any other SQLAlchemy session."""
        stateless_rows = []
        for row in rows:
            row_args = row.__dict__
            row_args.pop("_sa_instance_state", None)

            stateless_row = type(row)(**row_args)

            stateless_rows.append(stateless_row)
        return stateless_rows

    @staticmethod
    def _get_metadatas(rows):
        return {row.metadata for row in rows}

    @staticmethod
    def _create_rows(engine, rows):
        Session = sessionmaker(bind=engine)
        session = Session()

        session.add_all(rows)

        session.commit()
        session.close()


class Statements(AbstractAction):
    def __init__(self, *statements):
        self.statements = statements

    def run(self, engine_manager):
        for statement in self.statements:
            engine_manager.engine.execute(statement)


@attr.s
class EngineManager(object):
    engine = attr.ib()
    ordered_actions = attr.ib(default=attr.Factory(tuple))
    tables: Tuple = attr.ib(default=None, converter=attr.converters.optional(tuple))
    session = attr.ib(default=False)
    default_schema = attr.ib(default=None)

    _ddl_created = False

    def _run_actions(self):
        BaseType = type(declarative_base())

        for action in self.ordered_actions:
            if isinstance(action, MetaData):
                self.create_ddl(action)
            elif isinstance(action, BaseType):
                self.create_ddl(action.metadata)
            elif isinstance(action, AbstractAction):
                action.run(self)
            elif callable(action):
                self._execute_function(action)
            else:
                raise ValueError(
                    "create_fixture function takes in sqlalchemy.MetaData or actions as inputs only."
                )

    def _create_schemas(self, metadata):
        if self._ddl_created:
            return

        all_schemas = {table.schema for table in metadata.tables.values() if table.schema}

        for schema in all_schemas:
            if self.default_schema == schema:
                continue

            statement = CreateSchema(schema, quote=True)
            self.engine.execute(statement)

    def _create_tables(self, metadata):
        if not self.tables:
            metadata.create_all(self.engine)
            return

        table_objects = {
            table_object
            for table in self.tables
            for table_object in identify_matching_tables(metadata, table)
        }

        metadata.create_all(self.engine, tables=list(table_objects))

    def _execute_function(self, fn):
        Session = sessionmaker(bind=self.engine)
        session = Session()

        fn(session)

        session.commit()
        session.close()

    def create_ddl(self, metadata):
        self._create_schemas(metadata)
        self._create_tables(metadata)
        self._ddl_created = True

    def manage(self, session=None):
        try:
            self._run_actions()

            if session:
                if isinstance(session, sessionmaker):
                    session_factory = session
                else:
                    session_factory = sessionmaker(bind=self.engine)

                Session = scoped_session(session_factory)
                session = Session(bind=self.engine)
                yield session
                session.close()
            else:
                yield self.engine
        finally:
            self.engine.dispose()

    def manage_sync(self, session=None):
        try:
            self._run_actions()

            if session:
                if isinstance(session, sessionmaker):
                    session_factory = session
                else:
                    session_factory = sessionmaker(bind=self.engine)

                Session = scoped_session(session_factory)
                session = Session(bind=self.engine)
                yield session
                session.close()
            else:
                yield self.engine
        finally:
            self.engine.dispose()

    async def manage_async(self, session=None):
        try:
            self._run_actions()

            async_engine = self._get_async_engine()

            if session:
                if isinstance(session, sessionmaker):
                    session_factory = session
                else:
                    session_factory = sessionmaker(
                        async_engine,
                        expire_on_commit=False,
                        class_=compat.sqlalchemy.asyncio.AsyncSession,
                    )
                async with session_factory() as session:
                    yield session
            else:
                yield async_engine
        finally:
            self.engine.dispose()

    def _get_async_engine(self, isolation_level=None):
        url = compat.sqlalchemy.URL(
            drivername="postgresql+asyncpg",
            username=self.engine.pmr_credentials.username,
            password=self.engine.pmr_credentials.password,
            host=self.engine.pmr_credentials.host,
            port=self.engine.pmr_credentials.port,
            database=self.engine.pmr_credentials.database,
            query=dict(ssl="disable"),
        )
        options = {}
        if isolation_level:
            options["isolation_level"] = isolation_level
        return compat.sqlalchemy.asyncio.create_async_engine(url, **options)


def identify_matching_tables(metadata, table_specifier):
    if isinstance(table_specifier, DeclarativeMeta):
        return [table_specifier.__table__]

    if isinstance(table_specifier, Table):
        return [table_specifier]

    tables = [
        table
        for table_name, table in metadata.tables.items()
        if fnmatch.fnmatch(table_name, table_specifier)
    ]

    if tables:
        return tables

    table_names = ", ".join(sorted(metadata.tables.keys()))
    raise ValueError(
        'Could not identify any tables matching "{}" from: {}'.format(table_specifier, table_names)
    )



import binascii
import csv
import gzip
import io
import sys

from sqlalchemy import MetaData, Table

from pytest_mock_resources.compat import boto3


def execute_mock_s3_copy_command(statement, engine):
    params = _parse_s3_command(statement)

    _mock_s3_copy(
        table_name=params["table_name"],
        schema_name=params["schema_name"],
        s3_uri=params["s3_uri"],
        aws_secret_access_key=params["aws_secret_access_key"],
        aws_access_key_id=params["aws_access_key_id"],
        columns=params.get("columns", None),
        engine=engine,
    )


def _parse_s3_command(statement):
    """Format, Parse and call patched 'COPY' command."""
    statement = strip(statement)
    params = dict()

    # deleting copy
    tokens = statement.split()[1:]

    # Fetching table name
    params["schema_name"], params["table_name"] = _split_table_name(tokens.pop(0))

    # Checking for columns
    if tokens[0][0] == "(":
        ending_index = 0
        for index, arg in enumerate(tokens):
            if arg.endswith(")"):
                ending_index = index
                break

        ending_index += 1
        columns = tokens[0:ending_index]
        columns[0] = columns[0].replace("(", "")
        columns[-1] = columns[-1].replace(")", "")
        columns = [x.replace(",", "") for x in columns]
        columns = [x for x in columns if x != ""]
        tokens = tokens[ending_index:]
        params["columns"] = columns

    # Fetching s3_uri
    if tokens.pop(0).lower() != "from":
        raise ValueError(
            (
                "Possibly malformed S3 URI Format. "
                "Statement = {statement}"
                "Redshift fixture only supports S3 Copy statments with the following syntax: "
                "COPY <table_name> FROM [(column 1, [column2, [..]])] '<file path on S3 bucket>' "
                "credentials 'aws_access_key_id=<aws_access_key_id>;"
                "aws_secret_access_key=<aws_secret_access_key>'"
            ).format(statement=statement)
        )
    params["s3_uri"] = strip(tokens.pop(0))

    # Fetching credentials
    for token in tokens:
        if "aws_access_key_id" in token.lower() or "aws_secret_access_key" in token.lower():
            # This is because of the following possibiliteis:
            # ... [with ]credentials[ AS] 'aws_access_key_id=x;aws_secret_access_key=y'
            # OR
            # ... [with ]credentials[ AS] 'aws_secret_access_key=y;aws_access_key_id=x'
            # OR
            # ... [with ]credentials[ AS] 'aws_secret_access_key=y;\naws_access_key_id=x'
            # OR
            # ... [with ]credentials[ AS] 'aws_secret_access_key=y; aws_access_key_id=x'
            # Supportred AWS credentials format:
            # [with ]credentials[ AS] 'aws_secret_access_key=y; aws_access_key_id=x'
            # No Support for additional credential formats, eg IAM roles, etc, yet.
            credentials_list = token.split(";")
            for credentials in credentials_list:
                if "aws_access_key_id" in credentials:
                    params["aws_access_key_id"] = credentials.split("=")[-1]
                elif "aws_secret_access_key" in credentials:
                    params["aws_secret_access_key"] = credentials.split("=")[-1]
                else:
                    raise ValueError(
                        (
                            "Possibly malformed AWS Credentials Format. "
                            "Statement = {statement}"
                            "Redshift fixture only supports S3 Copy statments with the following "
                            "syntax: COPY <table_name> FROM [(column 1, [column2, [..]])] '"
                            "<file path on S3 bucket>' "
                            "credentials 'aws_access_key_id=<aws_access_key_id>;"
                            "aws_secret_access_key=<aws_secret_access_key>' "
                            "Supportred AWS credentials format: "
                            "[with ]credentials[ AS] 'aws_secret_access_key=y; aws_access_key_id=x'"
                            " No Support for additional credential formats, eg IAM roles, etc, yet."
                        ).format(statement=statement)
                    )
    return params


def _split_table_name(table_name):
    """Split 'schema_name.table_name' to (schema_name, table_name)."""
    table_name_items = table_name.split(".")
    if len(table_name_items) == 1:
        schema_name = None
    elif len(table_name_items) == 2:
        schema_name, table_name = table_name_items
    else:
        raise ValueError("Cannot determine schema/table name from input {}".format(table_name))
    return schema_name, table_name


def _mock_s3_copy(
    table_name, s3_uri, schema_name, aws_secret_access_key, aws_access_key_id, columns, engine
):
    """Execute patched 'copy' command."""
    s3 = boto3.client(
        "s3", aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key
    )
    ending_index = len(s3_uri)
    path_to_file = s3_uri[5:ending_index]
    bucket, key = path_to_file.split("/", 1)
    response = s3.get_object(Bucket=bucket, Key=key)

    # the following lins of code is used to check if the file is gzipped or not.
    # To do so we use magic numbers.
    # A mgic number is a constant numerical or text value used to identify a file format or protocol
    # The magic number for gzip compressed files is 1f 8b.
    is_gzipped = binascii.hexlify(response["Body"].read(2)) == b"1f8b"

    response = s3.get_object(Bucket=bucket, Key=key)
    data = read_data_csv(response["Body"].read(), is_gzipped, columns)

    meta = MetaData()
    table = Table(table_name, meta, autoload=True, schema=schema_name, autoload_with=engine)
    engine.execute(table.insert(data))


def read_data_csv(file, is_gzipped=False, columns=None, delimiter="|"):
    buffer = io.BytesIO(file)
    if is_gzipped:
        buffer = gzip.GzipFile(fileobj=buffer, mode="rb")

    # FUCK you python 2. This is ridiculous!
    wrapper = buffer
    if sys.version_info.major >= 3:
        wrapper = io.TextIOWrapper(buffer)
    else:
        delimiter = delimiter.encode("utf-8")

    reader = csv.DictReader(
        wrapper,
        delimiter=delimiter,
        quoting=csv.QUOTE_MINIMAL,
        quotechar='"',
        lineterminator="\n",
        skipinitialspace=True,
        doublequote=True,
    )
    return [dict(row) for row in reader]


def strip(input_string):
    """Strip trailing whitespace, single/double quotes."""
    return input_string.strip().rstrip(";").strip('"').strip("'")



class ImportAdaptor(object):
    __wrapped__ = False

    def __init__(self, package, recommended_extra, fail_message=None, **attrs):
        self.package = package
        self.recommended_extra = recommended_extra
        self.fail_message = fail_message

        for key, value in attrs.items():
            setattr(self, key, value)

    def fail(self):
        if self.fail_message:
            fail_message = self.fail_message
        else:
            fail_message = "Cannot use {recommended_extra} fixtures without {package}. pip install pytest-mock-resources[{recommended_extra}]".format(
                package=self.package, recommended_extra=self.recommended_extra
            )

        raise RuntimeError(fail_message)

    def __getattr__(self, attr):
        self.fail()



import contextlib
from unittest import mock

from sqlalchemy import create_engine
from sqlalchemy.sql.elements import TextClause

from pytest_mock_resources.compat import psycopg2
from pytest_mock_resources.patch.redshift.mock_s3_copy import strip
from pytest_mock_resources.patch.redshift.sqlalchemy import (
    execute_mock_s3_copy_command,
    execute_mock_s3_unload_command,
)


def mock_psycopg2_connect(config, _connect):
    """Patch `psycopg2._connect`.

    Add support for S3 COPY and UNLOAD.
    """

    class CustomCursor(psycopg2.extensions.cursor):
        """A custom cursor class to define a custom execute method."""

        def execute(self, sql, args=None):
            dsn_params = self.connection.get_dsn_parameters()
            engine = create_engine(
                "postgresql+psycopg2://{user}:{password}@{hostname}:{port}/{dbname}".format(
                    user=config.username,
                    password=config.password,
                    hostname=config.host,
                    port=config.port,
                    dbname=dsn_params["dbname"],
                )
            )
            if not isinstance(sql, TextClause) and strip(sql).lower().startswith("copy"):
                return execute_mock_s3_copy_command(sql, engine)
            if not isinstance(sql, TextClause) and strip(sql).lower().startswith("unload"):
                return execute_mock_s3_unload_command(sql, engine)
            return super(CustomCursor, self).execute(sql, args)

    def _mock_psycopg2_connect(*args, **kwargs):
        """Substitute the default cursor with a custom cursor."""
        conn = _connect(*args, **kwargs)
        conn.cursor_factory = CustomCursor
        return conn

    return _mock_psycopg2_connect


@contextlib.contextmanager
def patch_connect(config):
    new_connect = mock_psycopg2_connect(config, _connect=psycopg2._connect)
    with mock.patch("psycopg2._connect", new=new_connect) as p:
        yield p



import sqlalchemy.engine.url

from pytest_mock_resources.compat.import_ import ImportAdaptor

try:
    from sqlalchemy.ext import asyncio  # type: ignore
except ImportError:
    asyncio = ImportAdaptor(
        "SQLAlchemy",
        "SQLAlchemy >= 1.4",
        fail_message="Cannot use sqlalchemy async features with SQLAlchemy < 1.4.\n",
    )

URL = sqlalchemy.engine.url.URL
try:
    # Attempt to use the newly recommended `URL.create` method when available
    URL = URL.create  # type: ignore
except AttributeError:
    # But if it's not available, the top-level contructor should still exist,
    # as this is the only available option in sqlalchemy<1.4 versions.
    pass



import pytest
import sqlalchemy

from pytest_mock_resources.container.postgres import get_sqlalchemy_engine, PostgresConfig
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials
from pytest_mock_resources.fixture.database.relational.generic import EngineManager


@pytest.fixture(scope="session")
def pmr_postgres_config():
    """Override this fixture with a :class:`PostgresConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_postgres_config():
        ...     return PostgresConfig(image="postgres:9.6.10", root_database="foo")
    """
    return PostgresConfig()


def create_engine_manager(pmr_postgres_config, ordered_actions, tables):
    database_name = _create_clean_database(pmr_postgres_config)
    engine = get_sqlalchemy_engine(pmr_postgres_config, database_name)
    assign_fixture_credentials(
        engine,
        drivername="postgresql+psycopg2",
        host=pmr_postgres_config.host,
        port=pmr_postgres_config.port,
        database=database_name,
        username=pmr_postgres_config.username,
        password=pmr_postgres_config.password,
    )
    return EngineManager(engine, ordered_actions, tables=tables, default_schema="public")


def create_postgres_fixture(
    *ordered_actions, scope="function", tables=None, session=None, async_=False
):
    """Produce a Postgres fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
        async_: Whether to return an async fixture/client.
    """

    @pytest.fixture(scope=scope)
    def _sync(_postgres_container, pmr_postgres_config):
        engine_manager = create_engine_manager(pmr_postgres_config, ordered_actions, tables)
        yield from engine_manager.manage_sync(session=session)

    @pytest.fixture(scope=scope)
    async def _async(_postgres_container, pmr_postgres_config):
        engine_manager = create_engine_manager(pmr_postgres_config, ordered_actions, tables)
        async for engine in engine_manager.manage_async(session=session):
            yield engine

    if async_:
        return _async
    else:
        return _sync


def _create_clean_database(config):
    root_engine = get_sqlalchemy_engine(config, config.root_database, isolation_level="AUTOCOMMIT")

    try:
        root_engine.execute(
            """
            CREATE TABLE IF NOT EXISTS pytest_mock_resource_db(
                id serial
            );
            """
        )
    except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
        # A race condition may occur during table creation if:
        #  - another process has already created the table
        #  - the current process begins creating the table
        #  - the other process commits the table creation
        #  - the current process tries to commit the table creation
        pass

    result = root_engine.execute(
        "INSERT INTO pytest_mock_resource_db VALUES (DEFAULT) RETURNING id"
    )
    id_ = tuple(result)[0][0]
    database_name = "pytest_mock_resource_db_{}".format(id_)

    root_engine.execute('CREATE DATABASE "{}"'.format(database_name))
    root_engine.execute(
        'GRANT ALL PRIVILEGES ON DATABASE "{}" TO CURRENT_USER'.format(database_name)
    )

    return database_name



import pytest

from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials
from pytest_mock_resources.fixture.database.relational.generic import EngineManager
from pytest_mock_resources.fixture.database.relational.postgresql import (
    _create_clean_database,
    get_sqlalchemy_engine,
)
from pytest_mock_resources.patch.redshift import psycopg2, sqlalchemy


def create_redshift_fixture(*ordered_actions, scope="function", tables=None, session=None):
    """Produce a Redshift fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
    """

    from pytest_mock_resources.fixture.database.relational.redshift.udf import REDSHIFT_UDFS

    ordered_actions = ordered_actions + (REDSHIFT_UDFS,)

    @pytest.fixture(scope=scope)
    def _(_redshift_container, pmr_postgres_config):
        database_name = _create_clean_database(pmr_postgres_config)
        engine = get_sqlalchemy_engine(pmr_postgres_config, database_name)

        assign_fixture_credentials(
            engine,
            drivername="postgresql+psycopg2",
            host=pmr_postgres_config.host,
            port=pmr_postgres_config.port,
            database=database_name,
            username=pmr_postgres_config.username,
            password=pmr_postgres_config.password,
        )

        engine = sqlalchemy.substitute_execute_with_custom_execute(engine)
        engine_manager = EngineManager(
            engine, ordered_actions, tables=tables, default_schema="public"
        )

        with psycopg2.patch_connect(pmr_postgres_config):
            for engine in engine_manager.manage(session=session):
                yield engine

    return _



import functools
import os
import socket
from typing import Dict, Iterable

_DOCKER_HOST = "host.docker.internal"


def is_ci():
    return os.getenv("CI") == "true"


@functools.lru_cache()
def is_docker_host():
    try:
        socket.gethostbyname(_DOCKER_HOST)
    except socket.gaierror:
        return False
    else:
        return True


def get_env_config(name, kind):
    env_var = "PMR_{name}_{kind}".format(name=name.upper(), kind=kind.upper())
    return os.environ.get(env_var)


def fallback(fn):
    attr = fn.__name__

    @property
    @functools.wraps(fn)
    def wrapper(self):
        value = get_env_config(self.name, attr)
        if value is not None:
            return value

        value = getattr(self, "_{attr}".format(attr=attr), None)
        if value is not None:
            return value

        try:
            return fn(self)
        except NotImplementedError:
            if attr in self._fields_defaults:
                return self._fields_defaults[attr]
            return None

    return wrapper


class DockerContainerConfig:
    _fields: Iterable = {"image", "host", "port", "ci_port"}
    _fields_defaults: Dict = {}

    def __init__(self, **kwargs):
        for field in self._fields:
            value = kwargs.get(field)
            attr = "_{}".format(field)
            setattr(self, attr, value)

    def __repr__(self):
        cls_name = self.__class__.__name__
        return "{cls_name}({attrs})".format(
            cls_name=cls_name,
            attrs=", ".join(
                "{}={}".format(attr, repr(getattr(self, attr))) for attr in self._fields
            ),
        )

    @fallback
    def image(self):
        raise NotImplementedError()

    @fallback
    def ci_port(self):
        raise NotImplementedError()

    @fallback
    def host(self):
        if is_docker_host():
            return _DOCKER_HOST

        return os.environ.get("PYTEST_MOCK_RESOURCES_HOST", "localhost")

    @fallback
    def port(self):
        ci_port = self.ci_port
        if ci_port and is_ci():
            return ci_port

        raise NotImplementedError()



import pytest

from pytest_mock_resources.compat import redis
from pytest_mock_resources.config import DockerContainerConfig
from pytest_mock_resources.container.base import ContainerCheckFailed, get_container


class RedisConfig(DockerContainerConfig):
    """Define the configuration object for redis.

    Args:
        image (str): The docker image:tag specifier to use for redis containers.
            Defaults to :code:`"redis:5.0.7"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`6380`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`6379`.
    """

    name = "redis"

    _fields = {"image", "host", "port", "ci_port"}
    _fields_defaults = {
        "image": "redis:5.0.7",
        "port": 6380,
        "ci_port": 6379,
    }


def check_redis_fn(config):
    try:
        client = redis.Redis(host=config.host, port=config.port)
        client.ping()
    except redis.ConnectionError:
        raise ContainerCheckFailed(
            "Unable to connect to a presumed Redis test container via given config: {}".format(
                config
            )
        )


@pytest.fixture(scope="session")
def _redis_container(pmr_redis_config):
    result = get_container(
        pmr_redis_config,
        ports={6379: pmr_redis_config.port},
        environment={},
        check_fn=check_redis_fn,
    )
    yield next(iter(result))



from typing import Union

from sqlalchemy.sql.base import Executable

from pytest_mock_resources.compat import sqlparse
from pytest_mock_resources.patch.redshift.mock_s3_copy import execute_mock_s3_copy_command, strip
from pytest_mock_resources.patch.redshift.mock_s3_unload import execute_mock_s3_unload_command


def substitute_execute_with_custom_execute(engine):
    """Substitute the default execute method with a custom execute for copy and unload command."""
    default_execute = engine.execute

    def custom_execute(statement: Union[Executable, str], *args, **kwargs):
        if isinstance(statement, Executable):
            return default_execute(statement, *args, **kwargs)

        # The statement is assumed to be a string at this point.
        normalized_statement = strip(statement).lower()
        if normalized_statement.startswith("copy"):
            return execute_mock_s3_copy_command(statement, engine)
        elif normalized_statement.startswith("unload"):
            return execute_mock_s3_unload_command(statement, engine)

        return default_execute(statement, *args, **kwargs)

    def handle_multiple_statements(statement: Union[Executable, str], *args, **kwargs):
        """Split statement into individual sql statements and execute.

        Splits multiple statements by ';' and executes each.
        NOTE: Only the result of the last statements is returned.
        """
        statements_list = parse_multiple_statements(statement)
        result = None
        for statement in statements_list:
            result = custom_execute(statement, *args, **kwargs)

        return result

    # Now each statement is handled as if it contains multiple sql statements
    engine.execute = handle_multiple_statements
    return engine


def parse_multiple_statements(statement: Union[Executable, str]):
    """Split the given sql statement into a list of individual sql statements."""
    # Ignore SQLAlchemy Text Objects.
    if isinstance(statement, Executable):
        return [statement]

    # Preprocess input statement
    processed_statement = _preprocess(statement)
    statements_list = [str(statement) for statement in sqlparse.split(processed_statement)]

    return statements_list


def _preprocess(statement: str):
    """Preprocess the input statement."""
    statement = statement.strip()
    # Replace any occourance of " with '.
    statement = statement.replace('"', "'")
    if statement[-1] != ";":
        statement += ";"
    return statement



# flake8: noqa
from pytest_mock_resources.container import (
    _mongo_container,
    _mysql_container,
    _postgres_container,
    _redis_container,
    _redshift_container,
    MongoConfig,
    MysqlConfig,
    PostgresConfig,
    RedisConfig,
)
from pytest_mock_resources.fixture.database import (
    create_mongo_fixture,
    create_mysql_fixture,
    create_postgres_fixture,
    create_redis_fixture,
    create_redshift_fixture,
    create_sqlite_fixture,
    pmr_mongo_config,
    pmr_mysql_config,
    pmr_postgres_config,
    pmr_redis_config,
    Rows,
    Statements,
)
from pytest_mock_resources.hooks import pytest_configure, pytest_itemcollected



import pytest
import sqlalchemy

from pytest_mock_resources.container.mysql import get_sqlalchemy_engine, MysqlConfig
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials
from pytest_mock_resources.fixture.database.relational.generic import EngineManager


@pytest.fixture(scope="session")
def pmr_mysql_config():
    """Override this fixture with a :class:`MysqlConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_mysql_config():
        ...     return MysqlConfig(image="mysql:5.2", root_database="foo")
    """
    return MysqlConfig()


def create_mysql_fixture(*ordered_actions, scope="function", tables=None, session=None):
    """Produce a MySQL fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
    """

    @pytest.fixture(scope=scope)
    def _(_mysql_container, pmr_mysql_config):
        database_name = _create_clean_database(pmr_mysql_config)
        engine = get_sqlalchemy_engine(pmr_mysql_config, database_name)

        assign_fixture_credentials(
            engine,
            drivername="mysql+pymysql",
            host=pmr_mysql_config.host,
            port=pmr_mysql_config.port,
            database=database_name,
            username=pmr_mysql_config.username,
            password=pmr_mysql_config.password,
        )

        engine_manager = EngineManager(engine, ordered_actions, tables=tables)
        for engine in engine_manager.manage(session=session):
            yield engine

    return _


def _create_clean_database(config):
    root_engine = get_sqlalchemy_engine(config, config.root_database, isolation_level="AUTOCOMMIT")

    try:
        root_engine.execute(
            """
            CREATE TABLE IF NOT EXISTS pytest_mock_resource_db(
                id serial
            );
            """
        )
    except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
        # A race condition may occur during table creation if:
        #  - another process has already created the table
        #  - the current process begins creating the table
        #  - the other process commits the table creation
        #  - the current process tries to commit the table creation
        pass

    root_engine.execute("INSERT INTO pytest_mock_resource_db VALUES (DEFAULT)")
    result = root_engine.execute("SELECT LAST_INSERT_ID()")
    id_ = tuple(result)[0][0]
    database_name = "pytest_mock_resource_db_{}".format(id_)

    root_engine.execute("CREATE DATABASE {}".format(database_name))

    return database_name



from pytest_mock_resources.compat.import_ import ImportAdaptor

# isort: split
from pytest_mock_resources.compat import sqlalchemy  # noqa

try:
    import psycopg2
except ImportError:
    fail_message = (
        "Cannot use postgres/redshift fixtures without psycopg2.\n"
        "pip install pytest-mock-resources[postgres] or pytest-mock-resources[[postgres-binary].\n"
        "Additionally, pip install pytest-mock-resources[redshift] for redshift fixtures."
    )
    psycopg2 = ImportAdaptor(
        "psycopg2",
        "postgres",
        fail_message=fail_message,
        extensions=ImportAdaptor(
            "psycopg2", "psycopg2", fail_message=fail_message, cursor=ImportAdaptor
        ),
    )

try:
    import asyncpg
except ImportError:
    fail_message = (
        "Cannot use postgres async fixtures without asyncpg.\n"
        "pip install pytest-mock-resources[postgres-async].\n"
    )
    asyncpg = ImportAdaptor(
        "asyncpg",
        "postgres",
        fail_message=fail_message,
        extensions=ImportAdaptor(
            "asyncpg", "asyncpg", fail_message=fail_message, cursor=ImportAdaptor
        ),
    )

try:
    import boto3
except ImportError:
    boto3 = ImportAdaptor("boto3", "redshift")

try:
    import moto
except ImportError:
    moto = ImportAdaptor("moto", "redshift")

try:
    import sqlparse
except ImportError:
    sqlparse = ImportAdaptor(
        "sqlparse",
        "redshift",
    )

try:
    import pymongo
except ImportError:
    pymongo = ImportAdaptor("pymongo", "mongo")

try:
    import redis
except ImportError:
    redis = ImportAdaptor("redis", "redis")  # type: ignore

try:
    import pymysql
except ImportError:
    pymysql = ImportAdaptor("pymysql", "mysql")  # type: ignore
